from collections import defaultdict
from typing import Optional

import numpy as np

from doy.utils import smooth_conv, smooth_ema


class Logger:
    def __init__(self, use_wandb: bool = True):
        self.data = defaultdict(list)
        self.data_x = defaultdict(list)
        self.use_wandb = use_wandb

    def __call__(self, step: int, **kwargs):
        assert kwargs
        for k, v in list(kwargs.items()):
            if v is None:
                del kwargs[k]
                continue

            try:
                v = v.item()
            except AttributeError:
                pass

            self.data[k].append(v)
            self.data_x[k].append(step)

        if self.use_wandb:
            import wandb

            wandb.log(data=kwargs, step=step)

    def __getitem__(self, key):
        return np.array(self.data[key])

    def get(self, key, smooth_args: Optional[tuple] = ("ema", 0.9)):
        if smooth_args is None:
            return self[key]

        smoothing_method, smoothing_param = smooth_args
        if smoothing_method == "ema":
            return smooth_ema(self[key], smoothing_param)
        elif smoothing_method == "conv":
            return smooth_conv(self[key], smoothing_param)
        else:
            raise ValueError(
                f"Unknown smoothing method: {smoothing_method}, should be 'ema' or 'conv'."
            )

    @property
    def keys(self):
        return list(self.data.keys())

    # def asdict(self):
    #    return {"data": self.data, "data_x": self.data_x}



from typing import Iterable

import numpy as np
from matplotlib import pyplot as plt


def _plot_subplt(ax, d, auto_ylim):
    """ utility function for plot_grid """
    if isinstance(d[-1], str):
        *d, label = d
        ax.plot(*d, label=label)
        ax.legend()
    else:
        ax.plot(*d)
    q = 0.05
    auto_ylim = (
        # we use the previous bounds & the quantile (to discard outliers) & the most recent point (to avoid cutting it off)
        min(auto_ylim[0], np.quantile(d[-1], q), d[-1][-1] - np.std(d[-1]) * 0.08),
        max(auto_ylim[1], np.quantile(d[-1], 1 - q), d[-1][-1] + np.std(d[-1]) * 0.08),
    )
    return auto_ylim


def plot_grid(
    *data,
    nrows=None,
    ncols=None,
    titles=None,
    yscale=None,
    ylimits=None,
    axes_off=False,
    figsize=None,
    sharex=False,
    sharey=False,
    tight_layout=True
):
    """
    call this as plot_grid(d_1, d_2, ...)

    where d_i is a tuple containing 1-3 elements: ([xs], ys, [label])
    - (optional) a list xs of x values
    - a list ys of y values
    - (optional) a string used as legend label

    alternatively, d_i can also be a list of tuples as described above (for multiple plotted lines per subplots)
    """

    if nrows and ncols:
        assert nrows * ncols == len(data)
    elif nrows:
        assert len(data) % nrows == 0
        ncols = len(data) // nrows
    elif ncols:
        assert len(data) % ncols == 0
        nrows = len(data) // ncols
    else:
        nrows = 1
        ncols = len(data)

    if not figsize:
        k = 4
        figsize = (k * ncols, k * nrows)

    figs, axes = plt.subplots(
        nrows=nrows,
        ncols=ncols,
        squeeze=False,
        figsize=figsize,
        sharex=sharex,
        sharey=sharey,
    )

    for i, (d, ax) in enumerate(zip(data, axes.reshape(-1))):
        auto_ylim = (float("inf"), float("-inf"))
        if d is not None:
            if isinstance(d[0][0], Iterable):
                # if we have multiple plots in a subplot
                for k in d:
                    auto_ylim = _plot_subplt(ax, k, auto_ylim)
            else:
                auto_ylim = _plot_subplt(ax, d, auto_ylim)
        if axes_off:
            ax.axis("off")
        if titles:
            ax.set_title(titles[i])
        if yscale and yscale[i]:
            ax.set_yscale(yscale[i])
        if ylimits and ylimits[i] is not None:
            # ylimits[i] is (s1, s2) or s1 where s_i is either a scalar, or None, or 'auto' (latter one excludes outliers)

            if not isinstance(ylimits[i], tuple):
                ylimits[i] = (ylimits[i], ylimits[i])

            # to prevent: UserWarning: Attempt to set non-positive ylim on a log-scaled axis will be ignored.
            if yscale and yscale[i] == "log":
                auto_ylim = (max(auto_ylim[0], 0), max(auto_ylim[1], 0))

            # to prevent: UserWarning: Attempting to set identical low and high ylims makes transformation singular; automatically expanding.
            if auto_ylim[0] == auto_ylim[1]:
                auto_ylim = (None, None)

            ylimits[i] = list(ylimits[i])
            for j in range(2):
                if ylimits[i][j] == "auto":
                    ylimits[i][j] = auto_ylim[j]

            ax.set_ylim(bottom=ylimits[i][0], top=ylimits[i][1])

    if tight_layout:
        plt.tight_layout()
    plt.show()


def imshow_grid(
    *images, nrows=None, ncols=None, titles=None, axes_off=True, figsize=None
):
    if nrows and ncols:
        assert nrows * ncols == len(images)
    elif nrows:
        assert len(images) % nrows == 0
        ncols = len(images) // nrows
    elif ncols:
        assert len(images) % ncols == 0
        nrows = len(images) // ncols
    else:
        nrows = 1
        ncols = len(images)

    if not figsize:
        k = 12
        ratio = 1  # 0.25 if titles else 0.2
        figsize = (k * nrows, k * ncols)

    figs, axes = plt.subplots(nrows=nrows, ncols=ncols, squeeze=False, figsize=figsize)

    for i, (im, ax) in enumerate(zip(images, axes.reshape(-1))):
        if im is not None:
            ax.imshow(im, interpolation="nearest")
        if axes_off:
            ax.axis("off")
        if titles:
            ax.set_title(titles[i])

    plt.tight_layout()
    plt.show()



from numpy.compat import os_fspath
from numpy.lib.npyio import zipfile_factory
from numpy.lib.format import write_array


def savez_compressed(
    file, data_dict, compress_level=6, allow_pickle=True, pickle_kwargs=None
):
    """like np.savez_compressed but allows for variable compression level"""

    # Import is postponed to here since zipfile depends on gzip, an optional
    # component of the so-called standard library.
    import zipfile

    if not hasattr(file, "write"):
        file = os_fspath(file)
        if not file.endswith(".npz"):
            file = file + ".npz"

    if compress_level is not None:
        compression = zipfile.ZIP_DEFLATED
    else:
        compression = zipfile.ZIP_STORED

    assert (
        isinstance(compress_level, int) and compress_level >= 1 and compress_level <= 9
    )
    zipf = zipfile_factory(
        file, mode="w", compression=compression, compresslevel=compress_level
    )

    for key, val in data_dict.items():
        fname = key + ".npy"
        val = np.asanyarray(val)
        # always force zip64, gh-10776
        with zipf.open(fname, "w", force_zip64=True) as fid:
            write_array(
                fid, val, allow_pickle=allow_pickle, pickle_kwargs=pickle_kwargs
            )

    zipf.close()



# adapted from https://github.com/dsw/proquint/tree/master
import random


def _char_list_to_dict(char_list):
    return {c: k for (c, k) in zip(char_list, range(len(char_list)))}

UINT_TO_CONSONANT = "bdfghjklmnprstvz"
UINT_TO_VOWEL = "aiou"

CONSONANT_TO_UINT = _char_list_to_dict(UINT_TO_CONSONANT)
VOWEL_TO_UINT = _char_list_to_dict(UINT_TO_VOWEL)

MASK_LAST4 = 0xF
MASK_LAST2 = 0x3

CHARS_PER_CHUNK = 5

def _uint16_to_quint(uint16_val):
    val = uint16_val
    res = ["?"] * CHARS_PER_CHUNK
    for i in range(CHARS_PER_CHUNK):
        if i & 1:
            res[-i - 1] = UINT_TO_VOWEL[val & MASK_LAST2]
            val >>= 2
        else:
            res[-i - 1] = UINT_TO_CONSONANT[val & MASK_LAST4]
            val >>= 4
    return "".join(res)


def uint2quint(uint_val, separator="-"):
    """Convert 32-bit integer value into corresponding proquint string identifier.

    >>> uint2quint(0x7F000001, '-')
    lusab-babad

    :param uint_val: 32-bit integer value to encode
    :param separator: string to separate character quintets
    :return: proquint string identifier
    """
    if uint_val < 0 or uint_val > 0xFFFFFFFF:
        raise ValueError("uint_val should be in range 0-0xFFFFFFFF")
    return _uint16_to_quint(uint_val >> 16) + separator + _uint16_to_quint(uint_val)

def random_proquint(n=1, sep="-"):
    """ Generate a random proquint with n*16 bits of entropy."""
    res = [_uint16_to_quint(random.getrandbits(16)) for _ in range(n)]
    return sep.join(res)



import pprint
from pathlib import Path

from filelock import FileLock, Timeout

import doy.data as data

_kvstore_basepath = "~/.doy_kvstore"
_kvstore_path = Path(_kvstore_basepath + ".pickle").expanduser()
_kvstore_hr_path = Path(_kvstore_basepath + ".hr").expanduser()
_kvstore_lock_path = Path(_kvstore_basepath + ".lock").expanduser()
lock = FileLock(_kvstore_lock_path)


@lock
def _try_write_hr_store(store):
    try:
        s = pprint.pformat(store, indent=4, width=120)
    except Exception as e:
        s = f"An error occurred while writing the human-readable representation of the store: {e}"
    with open(_kvstore_hr_path, "w") as f:
        f.write(s)


@lock
def get_store() -> dict:
    return data.load(_kvstore_path, default={})


@lock
def set_store(store: dict):
    data.dump(store, _kvstore_path)
    _try_write_hr_store(store)


@lock
def get(key):
    store = get_store()
    return store[key]


@lock
def set(key, value):
    store = get_store()
    store[key] = value
    set_store(store)



import json
import os
import pickle
from pathlib import Path


def load(path: os.PathLike, default=None):
    path = Path(path)
    if not path.suffix in ('.pickle', '.json'):
        raise ValueError('path must end in either .pickle or .json.')

    if path.exists():
        if path.suffix == '.pickle':
            with open(path, 'rb') as f:
                return pickle.load(f)
        elif path.suffix == '.json':
            with open(path) as f:
                return json.load(f)
    elif default is not None:
        return default

    raise FileNotFoundError('File not found and default==None.')


def dump(value, path: os.PathLike):
    path = Path(path)
    if path.suffix == '.pickle':
        with open(path, 'wb') as f:
            pickle.dump(value, f)
    elif path.suffix == '.json':
        with open(path, 'w') as f:
            json.dump(value, f)
    else:
        raise ValueError('path must end in either .pickle or .json.')



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['doy']

package_data = \
{'': ['*']}

install_requires = \
['filelock>=3.12.2,<4.0.0',
 'matplotlib>=3.6.2',
 'numpy',
 'rich>=13.4.1',
 'tqdm>=4.65.0',
 'wandb>=0.15.5']

setup_kwargs = {
    'name': 'doy',
    'version': '1.9.1',
    'description': '',
    'long_description': '# Doy\n\nSimple utility package\n',
    'author': 'Dominik Schmidt',
    'author_email': 'schmidtdominik30@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8',
}


setup(**setup_kwargs)



from typing import *

from rich.progress import *


class RateColumn(ProgressColumn):
    """Renders human readable processing rate."""

    def render(self, task: "Task") -> Text:
        """Render the speed in iterations per second."""
        speed = task.finished_speed or task.speed
        if speed is None:
            return Text("", style="progress.percentage")
        return Text(f"{speed:.2f} it/s", style="progress.percentage")


class TimeElapsedColumn(ProgressColumn):
    """Renders time elapsed."""

    def __init__(
        self, compact: bool = False, table_column: Optional[Column] = None,
    ):
        self.compact = compact
        super().__init__(table_column=table_column)

    def render(self, task: "Task") -> Text:
        """Show time elapsed."""
        elapsed = task.finished_time if task.finished else task.elapsed
        if elapsed is None:
            return Text(
                "--:--" if self.compact else "-:--:--", style="progress.elapsed"
            )

        minutes, seconds = divmod(int(elapsed), 60)
        hours, minutes = divmod(minutes, 60)

        if self.compact and not hours:
            formatted = f"{minutes:02d}:{seconds:02d}"
        else:
            formatted = f"{hours:d}:{minutes:02d}:{seconds:02d}"

        return Text(formatted, style="progress.elapsed")


def track(
    sequence,
    description: str = "Working...",
    total: Optional[float] = None,
    auto_refresh: bool = True,
    console=None,
    transient: bool = False,
    refresh_per_second: float = 12.5,
    update_period: float = 0.1,
    disable: bool = False,
):
    columns = [
        # TaskProgressColumn(show_speed=True),
        # SpinnerColumn('dots10'),
        *(
            [TextColumn("[progress.description]{task.description}")]
            if description
            else []
        ),
        BarColumn(
            bar_width=40,
            style="bar.back",
            complete_style="bar.complete",
            finished_style="bar.finished",
            pulse_style="bar.pulse",
        ),
        MofNCompleteColumn(),
        RateColumn(),
        TimeRemainingColumn(compact=True),
        TimeElapsedColumn(compact=True),
    ]
    progress = Progress(
        *columns,
        auto_refresh=auto_refresh,
        console=console,
        transient=transient,
        refresh_per_second=refresh_per_second or 10,
        disable=disable,
    )

    with progress:
        yield from progress.track(
            sequence, total=total, description=description, update_period=update_period
        )



import itertools
from typing import Any, Dict, List, Tuple, Type, TypeVar, Union

import matplotlib.pyplot as plt
import numpy as np
import torch


def chunks(lst, n):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i : i + n]


def mappend(lists, values):
    for l, v in zip(lists, values):
        l.append(v)


def lerp(a, b, alpha=0.9):
    return alpha * a + (1 - alpha) * b


def smooth_ema(X, alpha=0.9):
    assert 0 <= alpha < 1
    if len(X) == 0:
        return X
    res = []
    z = X[0]
    for x in X:
        z = lerp(z, x, alpha)
        res.append(z)
    return np.array(res)


def smooth_conv(X, box_pts, mode="valid"):
    assert isinstance(box_pts, int)
    if len(X) == 0:
        return X
    box = np.ones(box_pts) / box_pts
    X_smooth = np.convolve(X, box, mode=mode)
    return X_smooth


def bchw_to_bhwc(x):
    assert len(x.shape) == 4
    if isinstance(x, np.ndarray):
        return x.transpose(0, 2, 3, 1)
    else:
        return x.permute(0, 2, 3, 1)


def bhwc_to_bchw(x):
    assert len(x.shape) == 4
    if isinstance(x, np.ndarray):
        return x.transpose(0, 3, 1, 2)
    else:
        return x.permute(0, 3, 1, 2)


def count_parameters(model, requires_grad_only=True):
    return sum(
        p.numel()
        for p in model.parameters()
        if p.requires_grad or not requires_grad_only
    )


def state_dict_orig(m):
    return m._orig_mod.state_dict() if hasattr(m, "_orig_mod") else m.state_dict()

def get_state_dicts(**m):
    return {k: state_dict_orig(v) for k, v in m.items()}

def get_params_from_modules(*modules):
    yield from itertools.chain(*[m.parameters() for m in modules])


class Schedule:
    def __call__(self, step: Union[int, np.ndarray, list]):
        """Returns the value of the schedule at the given step(s)"""
        raise NotImplementedError()

    @property
    def range(self) -> Any:
        """
        Returns the range of the schedule.
        This range is what .plot() plots, __call__ may accept values outside this range.
        """
        raise NotImplementedError()

    def plot(self):
        xs = np.arange(*self.range)
        plt.plot(xs, self(xs))
        plt.show()


class PiecewiseLinearSchedule(Schedule):
    def __init__(self, points, values):
        self.points = np.array(points)
        self.values = np.array(values)
        if not np.all(np.diff(points) > 0):
            raise ValueError("points must be monotonically increasing")
        if len(self.points) != len(self.values):
            raise ValueError("points and values need to be of the same length")

    @property
    def range(self):
        return self.points[0], self.points[-1]

    def __call__(self, step: Union[int, np.ndarray, list]):
        is_scalar = False
        if isinstance(step, int):
            step = np.array([step])
            is_scalar = True
        elif isinstance(step, list):
            step = np.array(step)

        if np.any(step < self.points[0]) or np.any(self.points[-1] < step):
            raise ValueError("t must be in the interval [points[0], points[-1]]")

        inds = np.searchsorted(self.points, step) - 1
        inds = np.clip(inds, 0, len(self.points) - 2)

        interp = (step - self.points[inds]) / (
            self.points[inds + 1] - self.points[inds]
        )
        result = self.values[inds] * (1 - interp) + self.values[inds + 1] * interp

        return result[0] if is_scalar else result

U = TypeVar('U', bound=torch.optim.Optimizer)

class LRScheduler:
    def __init__(
        self,
        opt_cls: Type[U],
        param_groups: Dict[str, Tuple[Schedule, List[torch.nn.Module]]],
    ):
        opt_param_groups = []
        self.schedules = []

        for name, (schedule, modules) in param_groups.items():
            opt_param_groups.append(
                {"params": get_params_from_modules(*modules), "lr": schedule(0)}
            )
            self.schedules.append((name, schedule))
        self.opt: U = opt_cls(params=opt_param_groups)

    @classmethod
    def make(
        cls,
        opt_cls: Type[U] = torch.optim.AdamW,
        **param_groups: Tuple[Schedule, List[torch.nn.Module]],
    ) -> tuple[U, "LRScheduler"]:
        scheduler = cls(opt_cls, param_groups)
        return scheduler.opt, scheduler

    def step(self, step):
        for pg, (_, schedule) in zip(self.opt.param_groups, self.schedules):
            pg["lr"] = schedule(step)

    def get_state(self):
        return {
            "lr_" + name: pg["lr"]
            for pg, (name, _) in zip(self.opt.param_groups, self.schedules)
        }


def normalize_into_range(lower, upper, v, ensure_in_range=False):
    assert lower <= upper
    if ensure_in_range and (v < lower or v > upper):
        raise ValueError(f"Value {v} is not in range [{lower}, {upper}]")

    return (v - lower) / (upper - lower)



import builtins
import time
from contextlib import contextmanager
from typing import Literal, Optional

import rich
import rich.console
import rich.progress
import rich.status
from tqdm.auto import tqdm, trange

from .rich_utils import track

_mode = None
_rich_console = None


def config_progress_backend(mode: Literal["rich", "tqdm", "fallback"] = "rich"):
    assert mode in ("rich", "tqdm", "fallback")
    global _mode
    _mode = mode
    if mode == "rich":
        global _rich_console
        _rich_console = rich.console.Console()


def _init_check():
    if _mode is None:
        config_progress_backend()


def print(*args, **kwargs):
    _init_check()
    if _mode == "rich":
        _rich_console.print(*args, **kwargs)
    else:
        builtins.print(*args, **kwargs)


def log(*args, **kwargs):
    _init_check()
    if _mode == "rich":
        _rich_console.log(*args, **kwargs)
    else:
        builtins.print(*args, **kwargs)


def loop(*args, desc: Optional[str] = None, desc_align: int = 25):
    _init_check()

    if desc is None:
        desc = ""
    else:
        desc = f"{desc:<{desc_align}}"

    if isinstance(args[0], int):
        iterable = range(*args)
    else:
        assert len(args) == 1
        iterable = args[0]

    if _mode == "tqdm":
        yield from tqdm(iterable, desc=desc)
    elif _mode == "fallback":
        if desc is None:
            desc = ""
        if hasattr(iterable, "__len__"):
            l = len(iterable)
            for i, x in enumerate(iterable):
                if i % max(l // 20, 1) == 0:
                    print(f"[{i}/{l} iters] {desc}")
                yield x
        else:
            last_print = 0
            for i, x in enumerate(iterable):
                if time.time() - last_print > 10:
                    print(f"[{i}/? iters] {desc}")
                    last_print = time.time()
                yield x
    elif _mode == "rich":
        yield from track(iterable, description=desc, console=_rich_console)


def status(name="Working..."):
    if _mode is None:
        config_progress_backend()

    if _mode == "rich":
        return rich.status.Status(
            name,
            console=_rich_console,
            spinner="dots",
            spinner_style="status.spinner",
            speed=1,
            refresh_per_second=12.5,
        )
    else:
        print(name)
        return contextmanager(lambda: iter([None]))()



from proquint import random_proquint, uint2quint

import doy.kv

from .data import dump, load
from .logger import Logger
from .plotting import imshow_grid, plot_grid
from .progress import log, loop, print, status
from .utils import *
